<?xml version="1.0" encoding="ISO-8859-1"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns="http://purl.oclc.org/dsdl/schematron"
             xmlns:clc="urn:oasis:names:tc:ubl:schema:CodeList-Constraints-0.5"
                version="1.0"
                exclude-result-prefixes="clc">
<!--
     An implementation of the UBL Code List Value Validation Methodology.

     A stylesheet to convert an agreed upon set of pointers to coded values
     for constraining instances of documents at particular XPath addresses.

     This stylesheet calls out to external template rules to accommodate the
     naming of metadata and the format of external code list enumerations,
     thus it is not hardwired to UBL or any external representation of codes.

     Implementation restriction:  at this time this script is checking
     individual code list association files for having non-conflicting prefixes
     for namespaces, but it is not checking the entire suite of prefixes for
     namespaces for every included file for unique usage of prefix values in
     the generated result.  This may go unreported and cause a problem in the
     Schematron processing.

  $Id: Constraints2Schematron-0.6.xsl,v 1.1 2006/06/18 16:56:04 G. Ken Holman Exp $

Universal Business Language (UBL)

Copyright (C) OASIS Open (2006). All Rights Reserved.

This document and translations of it may be copied and furnished to others, and
derivative works that comment on or otherwise explain it or assist in its
implementation may be prepared, copied, published and distributed, in whole or
in part, without restriction of any kind, provided that the above copyright
notice and this paragraph are included on all such copies and derivative works.
However, this document itself may not be modified in any way, such as by
removing the copyright notice or references to OASIS, except as needed for the
purpose of developing OASIS specifications, in which case the procedures for
copyrights defined in the OASIS Intellectual Property Rights document must be
followed, or as required to translate it into languages other than English.

The limited permissions granted above are perpetual and will not be revoked by
OASIS or its successors or assigns.

This document and the information contained herein is provided on an "AS IS"
basis and OASIS DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO ANY WARRANTY THAT THE USE OF THE INFORMATION HEREIN WILL NOT
INFRINGE ANY RIGHTS OR ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR
A PARTICULAR PURPOSE.

-->

<!--this shortcut makes some namespace work more succinct in the stylesheet-->
<xsl:variable name="clc-namespace" select="document('')/*/namespace::clc"/>

<!--this keeps code lists handy for reference-->
<xsl:key name="codelists" match="clc:CodeList" use="@xml:id"/>

<!--this remembers the association file root node for the initial file only-->
<xsl:variable name="assoc-file" select="document('',/)"/>
<xsl:variable name="assoc-uri"/>

<!--this makes the resulting file easy to edit-->
<xsl:output encoding="ISO-8859-1"/>

<xsl:template match="/">
  <xsl:param name="assoc-file" select="$assoc-file"/>
  <xsl:param name="assoc-uri" select="$assoc-uri"/>
  <!--prevent stylesheet abuse-->
  <xsl:if test="not($assoc-file/clc:CodeListConstraints)">
    <xsl:message terminate="yes">
      <xsl:if test="$assoc-uri">
        <xsl:text>Problem processing file '</xsl:text>
        <xsl:value-of select="$assoc-uri"/>
        <xsl:text>': </xsl:text>
      </xsl:if>
      <xsl:text>Expected the document element {</xsl:text>
      <xsl:value-of select="$clc-namespace"/>
      <xsl:text>}CodeListConstraints, but instead have {</xsl:text>
      <xsl:value-of select="concat( namespace-uri(*),'}',local-name(*) )"/>
    </xsl:message>
  </xsl:if>
  <xsl:if test="not($assoc-uri)">
    <!--first time around, check the include tree for infinite loops-->
    <xsl:call-template name="check-infinite-loops"/>
  </xsl:if>
  <!--check given instance for inconsistencies-->
  <xsl:call-template name="check-inconsistencies"/>
  <xsl:choose>
    <xsl:when test="$assoc-uri">
      <!--this has been invoked by an Include directive, so just add rules-->
      <xsl:call-template name="add-rules">
        <xsl:with-param name="assoc-file" select="$assoc-file"/>
        <xsl:with-param name="assoc-uri" select="$assoc-uri"/>
      </xsl:call-template>
    </xsl:when>
    <xsl:otherwise>
      <!--create the pattern-->
      <pattern id="{$assoc-file/clc:CodeListConstraints/@name}">
        <xsl:call-template name="add-rules">
          <xsl:with-param name="assoc-file" select="$assoc-file"/>
          <xsl:with-param name="assoc-uri" select="$assoc-uri"/>
        </xsl:call-template>
      </pattern>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!--========================================================================-->
<!--add rules to the result-->

<!--add the rules from the context association file to the result-->
<xsl:template name="add-rules">
  <xsl:param name="assoc-file" select="$assoc-file"/>
  <xsl:param name="assoc-uri" select="$assoc-uri"/>
  <xsl:text>
</xsl:text>
  <xsl:if test="not($assoc-uri)">
    <!--$comment must be supplied by an external fragment; not defined here-->
    <xsl:comment><xsl:value-of select="$comment"/></xsl:comment><xsl:text>
</xsl:text>
  </xsl:if>
  <xsl:comment>
    Start of synthesis of rules from code list context associations.
<!--this will help identify the sets of rules-->
<xsl:copy-of select="$assoc-file/*/clc:Identification/node()"/>
<xsl:text>
</xsl:text>
<xsl:variable name="ns" select="$assoc-file//namespace::*"/>
<xsl:if test="$ns[name(.) and name(.)!='xml']">
    Required namespace declarations as indicated in this set of rules:

<xsl:for-each select="$ns[name(.) and name(.)!='xml']">
  <xsl:if test="generate-id(.)=generate-id($ns[name(.)=name(current())][1])">
    <xsl:value-of select="concat('&lt;ns prefix=&quot;',name(.),
                          '&quot; uri=&quot;',.,'&quot;/>')"/>
      <xsl:text>
</xsl:text>
  </xsl:if>
</xsl:for-each>
      <xsl:text>
</xsl:text>
</xsl:if>
</xsl:comment><xsl:text>
</xsl:text>
  <xsl:apply-templates select="$assoc-file/*/clc:Contexts/clc:Context">
    <xsl:with-param name="assoc-uri" select="$assoc-uri"/>
  </xsl:apply-templates>
  <xsl:comment>
    End of synthesis of rules from code list context associations.
</xsl:comment><xsl:text>
</xsl:text>
  <!--copy any indentation that was at the end of the rules-->
  <xsl:copy-of select="node()[last()][self::text()]"/>

  <xsl:for-each select="$assoc-file/*/clc:Include">
    <xsl:sort select="position()" order="descending" data-type="number"/>
    <xsl:call-template name="add-rules">
      <xsl:with-param name="assoc-file" select="document(@uri,.)"/>
      <xsl:with-param name="assoc-uri" select="@uri"/>
    </xsl:call-template>
  </xsl:for-each>
</xsl:template>

<!--========================================================================-->
<!--create the rules; one rule for each context-->

<xsl:template match="clc:Context" name="gather-context">
  <xsl:param name="assoc-uri"/>
  <xsl:param name="codelist-references" 
             select="concat( normalize-space(@codes), ' ' )"/>
  <xsl:param name="codelists-referenced" select="/.."/>
  <xsl:choose>
    <xsl:when test="contains( $codelist-references,' ' )">
      <!--gather up code list references for the given context-->
      <xsl:call-template name="gather-context">
        <xsl:with-param  name="codelist-references"
                         select="substring-after($codelist-references,' ')"/>
        <xsl:with-param  name="codelists-referenced"
                         select="$codelists-referenced | 
              key( 'codelists', substring-before($codelist-references,' ') )"/>
        <xsl:with-param name="assoc-uri" select="$assoc-uri"/>
      </xsl:call-template>
    </xsl:when>
    <xsl:otherwise>
      <!--now have the union set of all referenced in given context-->
      <xsl:variable name="context">
        <!--assume the context is the item address, but xpath overrides it-->
        <xsl:choose>
          <xsl:when test="@xpath"><xsl:value-of select="@xpath"/></xsl:when>
          <xsl:when test="@context">
            <xsl:value-of select="concat(@context,'//',@item)"/>
          </xsl:when>
          <xsl:otherwise><xsl:value-of select="@item"/></xsl:otherwise>
        </xsl:choose>
      </xsl:variable>
      <!--make the rules for this context-->
      <xsl:call-template name="code-list-make-rule">
        <xsl:with-param name="item" select="normalize-space(@item)"/>
        <xsl:with-param name="context" select="$context"/>
        <xsl:with-param name="value" select="normalize-space(@value)"/>
        <xsl:with-param name="codelists-referenced" 
                        select="$codelists-referenced"/>
        <xsl:with-param name="codes-identifier" 
                        select="normalize-space(@codes)"/>
        <xsl:with-param name="assoc-uri" select="$assoc-uri"/>
      </xsl:call-template>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!--========================================================================-->
<!--consider namespaces in the XPath addresses for Schematron completeness-->

<!--check given constraint file for duplicate namespace prefix use (Schematron
    file can only have one binding to any given prefix value)-->
<!--NOTE: this does not check uniqueness across suite of included files-->
<xsl:template name="declare-namespaces">
  <xsl:param name="indent"/>
  <xsl:param name="assoc-file" select="$assoc-file"/>
  <xsl:variable name="namespaces" select="$assoc-file//namespace::*"/>
  <xsl:variable name="check-namespaces-for-duplicates">
    <xsl:for-each select="$namespaces">
      <xsl:if test="generate-id(.)=
                    generate-id($namespaces[name(.)=name(current())][1])">
        <!--report a problem-->
        <xsl:if test=". != $namespaces[name(.)=name(current())]">
          <xsl:if test="$assoc-uri">
            <xsl:text>Problem processing file '</xsl:text>
            <xsl:value-of select="$assoc-uri"/>
            <xsl:text>': </xsl:text>
          </xsl:if>
          <xsl:text>Namespace prefix "</xsl:text>
          <xsl:value-of select="name(.)"/>
          <xsl:text>" has multiple URI values in the instance, </xsl:text>
          <xsl:text>and can only have one as a limitation of Schematron:
</xsl:text>
          <!--report all places where conflicting URIs are declared-->
          <xsl:variable name="uris" 
                        select="$namespaces[name(.)=name(current())]"/>
          <xsl:for-each select="$uris">
            <xsl:if test="generate-id(.)=generate-id($uris[.=current()][1])">
              <xsl:text>  </xsl:text>
              <xsl:call-template name="report-current-node-context"/>
              <xsl:text/>=<xsl:value-of select="."/>
              <xsl:text>
</xsl:text>
            </xsl:if>
          </xsl:for-each>
        </xsl:if>
      </xsl:if>
    </xsl:for-each>
  </xsl:variable>
  <!--require namespace prefixes to be unique across the entire document-->
  <xsl:if test="string($check-namespaces-for-duplicates)">
    <xsl:message terminate="yes">
      <xsl:value-of select="$check-namespaces-for-duplicates"/>
    </xsl:message>
  </xsl:if>
  <!--declare namespaces assumed to be used for XPath addresses-->
  <xsl:for-each select="$namespaces">
    <xsl:if test="generate-id(.)=
                  generate-id($namespaces[name(.)=name(current())][1]) and
                  name(.) != 'xml' and . != $clc-namespace" >
      <ns uri="{.}" prefix="{name(.)}"/>
      <xsl:copy-of select="$indent"/>
    </xsl:if>
  </xsl:for-each>
  <!--declare namespaces for each of the included files-->
  <xsl:for-each select="$assoc-file/*/clc:Include">
    <xsl:sort select="position()" order="descending" data-type="number"/>
    <xsl:call-template name="declare-namespaces">
      <xsl:with-param name="assoc-file" select="document(@uri,.)"/>
      <xsl:with-param name="assoc-uri" select="@uri"/>
      <xsl:with-param name="indent" select="$indent"/>
    </xsl:call-template>
  </xsl:for-each>
</xsl:template>

<!--========================================================================-->
<!--infinite loop check for include tree-->

<xsl:template name="check-infinite-loops">
  <xsl:variable name="bad-loop">
    <!--generate any messages-->
    <xsl:call-template name="check-given-include-tree"/>
  </xsl:variable>

  <!--exit if messages-->
  <xsl:if test="string($bad-loop)">
    <xsl:message terminate="yes">
      <xsl:copy-of select="$bad-loop"/>
    </xsl:message>
  </xsl:if>
</xsl:template>

<!--recurse through nested include directives-->
<xsl:template name="check-given-include-tree">
  <xsl:param name="root-node-ids" select="' '"/>
  <xsl:param name="uri" select="'{invocation document}'"/>
  <xsl:param name="last-uri"/>
  <xsl:variable name="this-root-node-id" select="generate-id(document('',/))"/>
  <xsl:choose>
    <xsl:when test="contains( $root-node-ids,
                              concat(' ',$this-root-node-id,' ') )">
      <!--we've ended up pointing to somewhere in our ancestry-->
      <xsl:text>Infinite loop of include statements triggered by </xsl:text>
      <xsl:text/>including '<xsl:value-of select="$uri"/>' from '<xsl:text/>
      <xsl:value-of select="$last-uri"/>'<xsl:text>
</xsl:text>
    </xsl:when>
    <xsl:otherwise>
      <!--so far so good, now check down next level of children-->
      <xsl:variable name="past-root-node-ids"
                   select="concat( $root-node-ids, $this-root-node-id, ' ' )"/>
      <xsl:for-each select="document('',/)/*/clc:Include">
        <xsl:variable name="new-uri" select="@uri"/>
        <xsl:for-each select="document($new-uri,.)">
          <xsl:call-template name="check-given-include-tree">
            <xsl:with-param name="root-node-ids" select="$past-root-node-ids"/>
            <xsl:with-param name="uri" select="$new-uri"/>
            <xsl:with-param name="last-uri" select="$uri"/>
          </xsl:call-template>
        </xsl:for-each>
      </xsl:for-each>
    </xsl:otherwise>
  </xsl:choose>
</xsl:template>

<!--========================================================================-->
<!--consistency checks for user's expression of constraints-->

<xsl:template name="check-inconsistencies">
  <xsl:variable name="inconsistencies">
    <!--good refs-->
    <xsl:for-each select="/*/clc:Contexts/
                          clc:Context[normalize-space(@codes)='']">
      <xsl:text>A context cannot have no references to codes: </xsl:text>
      <xsl:call-template name="report-current-node-context"/>
      <xsl:text>
</xsl:text>
    </xsl:for-each>

    <!--defined ids-->
    <xsl:for-each select="/*/clc:CodeList[normalize-space(@xml:id)='']">
      <xsl:text>A code list cannot be listed without an xml:id </xsl:text>
      <xsl:text>unique identifier: </xsl:text>
      <xsl:call-template name="report-current-node-context"/>
      <xsl:text>
</xsl:text>
    </xsl:for-each>

    <!--no duplicate ids-->
   <xsl:for-each select="//clc:CodeList[count(key('codelists',@xml:id))>1][1]">
    <xsl:text>A code list xml:id identifier must be unique (xml:id="</xsl:text>
      <xsl:value-of select="@xml:id"/>
      <xsl:text>"): 
</xsl:text>
      <xsl:for-each select="key('codelists',@xml:id)">
        <xsl:text>  </xsl:text>
        <xsl:call-template name="report-current-node-context"/>
        <xsl:text>
</xsl:text>
      </xsl:for-each>
    </xsl:for-each>

    <!--no missing or conflicting contexts-->
    <xsl:for-each select="/*/clc:Contexts/clc:Context">
      <xsl:choose>
        <xsl:when test="@xpath and @context">
          <xsl:text>A context cannot have both @xpath and @context: </xsl:text>
          <xsl:call-template name="report-current-node-context"/>
          <xsl:text>
</xsl:text>
        </xsl:when>
      </xsl:choose>
    </xsl:for-each>

    <!--check all document references to point to genericode instances-->
    <xsl:for-each select="/*/clc:CodeLists/clc:CodeList">
      <xsl:variable name="doc" select="document(@uri,.)"/>
      <xsl:choose>
        <xsl:when test="not($doc)">
        <xsl:text>Resource not found or cannot be processed as XML: </xsl:text>
          <xsl:value-of select="concat(@uri,'&#xa;  ')"/>
          <xsl:call-template name="report-current-node-context"/>
          <xsl:text>
</xsl:text>
        </xsl:when>
        <xsl:otherwise>
          <xsl:call-template name="code-list-confirm-instance">
            <xsl:with-param name="uri" select="@uri"/>
            <xsl:with-param name="doc" select="$doc"/>
          </xsl:call-template>
        </xsl:otherwise>
      </xsl:choose>
    </xsl:for-each>

  </xsl:variable>

  <xsl:if test="string($inconsistencies)">
    <xsl:message terminate="yes">
      <xsl:value-of select="$inconsistencies"/>
    </xsl:message>
  </xsl:if>
</xsl:template>

<!--========================================================================-->
<!--string manipulation for comparisons-->

<xsl:template name="build-metadata-comparison-string">
  <xsl:param name="item"/>

  <!--build a string of expressions checking for meta data-->
  <xsl:variable name="properties">
    <xsl:call-template name="document-model-metadata">
      <xsl:with-param name="item" select="$item"/>
    </xsl:call-template>
  </xsl:variable>

  <!--the first " and " in the string is not required, so trim it off-->
  <xsl:value-of select="substring($properties,6)"/>
</xsl:template>

<!--express the boolean expression of a property's value needing to be set;
    note by using a double negative of not(x!=y) instead of (x=y) is required
    so that the test will not fail if an operand is not present.  Using a
    double negative will return true if the operand is not present, and will
    return false if the operand is present but the wrong value.-->
<xsl:template name="check-list-property">
  <xsl:param name="address"/>
  <xsl:param name="property"/>
  <xsl:variable name="code-list-property-value">
    <xsl:call-template name="code-list-get-property-value">
      <xsl:with-param name="property" select="$property"/>
    </xsl:call-template>
  </xsl:variable>
  <xsl:if test="string($code-list-property-value)">
    <xsl:value-of select='concat(" and not(",$address,
                          "!=&apos;",$code-list-property-value,"&apos;)")'/>
  </xsl:if>
</xsl:template>

<!--========================================================================-->
<!--diagnostic report-->

<xsl:template name="report-current-node-context">
  <!--start with the XPath address of the element in the context expression-->
  <xsl:for-each select="ancestor-or-self::*">
    <xsl:text/>/<xsl:value-of select="name(.)"/>
    <xsl:variable name="preceding-count" 
                select="count(preceding-sibling::*[name(.)=name(current())])"/>
    <xsl:if test="$preceding-count > 0">
      <xsl:text/>[<xsl:value-of select="$preceding-count + 1"/>]<xsl:text/>
    </xsl:if>
  </xsl:for-each>
  <!--suffix with attribute or namespace node as appropriate-->
  <xsl:if test="not(self::*)">
    <xsl:text>/</xsl:text>
    <xsl:if test="count(../@*) = count(.|../@*)">@</xsl:if>
    <xsl:if test="count(../namespace::*) = count(.|../namespace::*)">
      <xsl:text>namespace::</xsl:text>
    </xsl:if>
    <xsl:value-of select="name(.)"/>
  </xsl:if>
</xsl:template>

</xsl:stylesheet>